/*
 * types.c
 * 
 * Copyright (c) 2000, BindView Corporation.
 *
 * See LICENSE file.
 *
 */


#include "syscalls.h"
#include <ntddk.h>
#include "ioctlcmd.h"

#define MagicFoo(type) extern void copy_arg_##type (void *, DWORD);
#include "typesx.h"
#undef MagicFoo

struct type_info all_types[] = {
#define MagicFoo(type) { copy_arg_##type },
#include "typesx.h"
#undef MagicFoo
};

static unsigned long copy_buff_safe (void *dest, void *src, size_t sz);

void
copy_arg_DWORD (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;

    ent->args[ent->args_size/4] = arg;
    ent->args_size += 4;
}

void
copy_arg_PDWORD (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned long rc;
    unsigned long val;

    rc = copy_buff_safe (&val, (void *)arg, 4);
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
    } else {
        ent->args[ent->args_size/4] = val;
    }
    ent->args_size += 4;
}


void
copy_arg_HANDLE (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;

    ent->args[ent->args_size/4] = arg;
    ent->args_size += 4;
}

void
copy_arg_PHANDLE (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned long rc;
    unsigned long val;

    rc = copy_buff_safe (&val, (void *)arg, 4);
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
    } else {
        ent->args[ent->args_size/4] = val;
    }
    ent->args_size += 4;
}


void
copy_arg_POBJ_ATTRIB (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned char *buff;
    unsigned long rc;
    DWORD need;
    OBJECT_ATTRIBUTES attr;
    UNICODE_STRING str;
    POBJECT_ATTRIBUTES pObj = (POBJECT_ATTRIBUTES) arg;

    rc = copy_buff_safe (&attr, (void *)arg, sizeof (attr));
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
        ent->args_size += 4;
    } else {
        memcpy (&ent->args[ent->args_size/4], &attr, sizeof (attr));
        ent->args_size += sizeof (attr);

        rc = copy_buff_safe (&str, (void *)attr.ObjectName, sizeof (*attr.ObjectName));
        ent->args[ent->args_size/4] = rc;
        ent->args_size += 4;
        if (rc == -1) {
            ent->args[ent->args_size/4] = (unsigned long)attr.ObjectName;
            ent->args_size += 4;
        } else {
            memcpy (&ent->args[ent->args_size/4], &str, sizeof (str));
            ent->args_size += sizeof (str);

            rc = copy_buff_safe (&ent->args[ent->args_size/4+1],
                                 str.Buffer, str.Length);
            ent->args[ent->args_size/4] = rc;
            ent->args_size += 4;
            if (rc == -1) {
                ent->args[ent->args_size/4] = (unsigned long)str.Buffer;
                ent->args_size += 4;
            } else {
                /* be sure it's null terminated as courtesy to user space */
                if (((unsigned short *)(&ent->args[ent->args_size/4]))[str.Length/2 - 1]
                    != L'\0') {
                    ((unsigned short *)&ent->args[ent->args_size/4])[str.Length/2] = L'\0';
                    str.Length += 2;
                }
                ent->args_size += ((str.Length + 3) / 4) * 4;
            }
        }
    }
#if 0    
    buff = (unsigned char *)&ent->args[ent->args_size/4];

    /* figure out how big what we're copying is */
    need = 4;
    if (pObj) {
        need += 8 + (pObj->ObjectName ? (pObj->ObjectName->Length) : 0) + 2;
    }
    need = ((need + 3) / 4) * 4;

    /* FIXME: verify it fits */
    
    /* put entry of size in args */
    *(DWORD *)buff = need;
    buff += sizeof (DWORD);

    if (pObj) {
        /* put thing itself in args */
        *(HANDLE *)buff = pObj->RootDirectory;
        buff += sizeof (HANDLE);

        *(DWORD *)buff = pObj->Attributes;
        buff += sizeof (DWORD);

        if (pObj->ObjectName) {
            memcpy (buff, pObj->ObjectName->Buffer, pObj->ObjectName->Length);
            buff += pObj->ObjectName->Length;
        }
        *(unsigned short *)buff = 0;
        buff += sizeof (unsigned short);
    }

    ent->args_size += need;
#endif
}


void
copy_arg_PUNICODE_STRING (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned char *buff;
    unsigned long rc;
    DWORD need;
    UNICODE_STRING str;

    rc = copy_buff_safe (&str, (void *)arg, sizeof (str));
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
        ent->args_size += 4;
    } else {
        memcpy (&ent->args[ent->args_size/4], &str, sizeof (str));
        ent->args_size += sizeof (str);

        rc = copy_buff_safe (&ent->args[ent->args_size/4+1],
                             str.Buffer, str.Length);
        ent->args[ent->args_size/4] = rc;
        ent->args_size += 4;
        if (rc == -1) {
            ent->args[ent->args_size/4] = (unsigned long)str.Buffer;
            ent->args_size += 4;
        } else {
            /* be sure it's null terminated as courtesy to user space */
            if (((unsigned short *)(&ent->args[ent->args_size/4]))[str.Length/2 - 1]
                != L'\0') {
                ((unsigned short *)&ent->args[ent->args_size/4])[str.Length/2] = L'\0';
                str.Length += 2;
            }
            ent->args_size += ((str.Length + 3) / 4) * 4;
        }
    }
}

void
copy_arg_ACCESS_MASK (void *e, DWORD arg)
{
    copy_arg_DWORD (e, arg);
}

void
copy_arg_PCLIENT_ID (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned long rc;
    unsigned char cid[8];

    rc = copy_buff_safe (&cid, (void *)arg, sizeof (cid));
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
        ent->args_size += 4;
    } else {
        memcpy (&ent->args[ent->args_size/4], &cid, sizeof (cid));
        ent->args_size += sizeof (cid);
    }
}

void
copy_arg_PLPC_MESSAGE (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned long rc;
    unsigned char msg[24]; /* actually may be bigger, but ignore that now */

    rc = copy_buff_safe (&msg, (void *)arg, sizeof (msg));
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
        ent->args_size += 4;
    } else {
        memcpy (&ent->args[ent->args_size/4], &msg, sizeof (msg));
        ent->args_size += sizeof (msg);
    }
}


void
copy_arg_PLPC_SECTIONINFO (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned long rc;
    unsigned char msg[24];

    rc = copy_buff_safe (&msg, (void *)arg, sizeof (msg));
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
        ent->args_size += 4;
    } else {
        memcpy (&ent->args[ent->args_size/4], &msg, sizeof (msg));
        ent->args_size += sizeof (msg);
    }
}


void
copy_arg_PLPC_SECTIONMAPINFO (void *e, DWORD arg)
{
    ENTRY *ent = (ENTRY *)e;
    unsigned long rc;
    unsigned char msg[12]; /* actually may be bigger, but ignore that now */

    rc = copy_buff_safe (&msg, (void *)arg, sizeof (msg));
    ent->args[ent->args_size/4] = rc;
    ent->args_size += 4;
    if (rc == -1) {
        ent->args[ent->args_size/4] = arg;
        ent->args_size += 4;
    } else {
        memcpy (&ent->args[ent->args_size/4], &msg, sizeof (msg));
        ent->args_size += sizeof (msg);
    }
}

static unsigned long probe_addr = 0x7ffef000;

static unsigned long
copy_buff_safe (void *dest, void *src, size_t sz)
{
    unsigned long rc = 0;

    /*
     * Technically, this == 0 check isn't necessary.  The __try block
     * below will catch such cases, and recover fine.  Problem is that
     * when the exception handler is invoked, it ends up calling
     * NtContinue, which is then seen by our hook and then user space.
     * So, to avoid confusing the user, try to avoid that.
     *
     * Note that that may still happen if the pointer passed is non-null
     * but still invalid.  Nothing we can do about that.  The null case
     * is very common though, and was quite distracting.  Took me a
     * while to figure out wtf was going on.  :)
     */
    if (src == 0)
        return -1;

    /*
     * FIXME: only do the following check if the call originates from
     * user-space.  Some kernel-space calls are failing this check,
     * and we lose their information, unnecessarily.
     */

    /* test userprobeaddress */
    if ((ExGetPreviousMode () == 1)
        && (((unsigned long) src > probe_addr)
            || ((unsigned long) src + sz > probe_addr))) {
        return -1;
    }

    __try {
        memcpy (dest, src, sz);
    } __except (EXCEPTION_EXECUTE_HANDLER) {
        rc = -1;
    }
    return rc;
}
